'''
Fourier Perturbation with Gaussian Noise
Author: Gergely Acs <gergely.acs@inria.fr>

Note: As Gaussian perturbation requires the L2-sensitivity of the data,
and the orthonormal DFT preserves L2-norm, it can result in larger accuracy
than Laplace (at the cost of weaker privacy). This gain is more significant if 
larger number of coefficients are kept (i.e., the data is less compressible).
'''

from math import *
from Utils import *
import ExpMechanism
from ExpMechanism import *
import numpy as np
import cmath

def EFPAG(data, l2_sens, eps, delta, debug = False):
    data = map(float, data)
    coeffs = np.array(rfft(data))

    n = len(coeffs)
    N = len(data)

    # DFT in scipy is not orthonormal
    coeffs /= sqrt(N)

    eps_1 = eps_2 = eps * 0.5
    sigma = l2_sens * sqrt(2.0 * log(4.0/delta))/eps_2

    if debug:
        print "L2-sensitivity:", l2_sens
        print "Epsilon (noise)", eps_2
        print "Sigma:", sigma
        print "Delta", delta

    s = abs(coeffs[0])**2 
    for i in range(1,n-1):
        s += 2*abs(coeffs[i])**2 

    if len(data) % 2 == 0:
        s += abs(coeffs[n-1])**2 
    else:
        s += 2*abs(coeffs[n-1])**2 
    
    # DC coeff
    kept = 1
    d = abs(coeffs[0])**2
    sum = sqrt(s - d) + sqrt(2*kept) * sigma
    #sum = sqrt(sum) * sqrt(kept)
    priv_items = [PrivItem(-sum, [0, kept])]
    
    # Other coeffs except the last one
    for i in xrange(1, n-1):
        kept += 2
        d += 2 * abs(coeffs[i])**2
        sum = sqrt(s - d) + sqrt(2*kept) * sigma
        #sum = sqrt(sum) * sqrt(kept)
        priv_items.append(PrivItem(-sum, [i, kept]))

    # last coeff needs special consideration depending len(data) is even or odd
    if len(data) % 2 == 0:
        kept += 1
        d += abs(coeffs[n-1])**2
    else:
        kept += 2
        d += 2 * abs(coeffs[n-1])**2

    sum = sqrt(s - d) + sqrt(2*kept) * sigma
    #sum = sqrt(sum) * sqrt(kept)
    priv_items.append(PrivItem(-sum, [n-1, kept]))

    # maximal kept coeffs is upper bounded by the length data vector
    item = ExpMechanism.run(priv_items, eps_1, l2_sens)

    k = item.id[0] + 1

    if debug:
        print "Kept coefficients: %d / %d" % (k, len(data))

    for j in range(n):
        if j < k:
            coeffs[j] += np.random.normal(0,sigma) + 1j * np.random.normal(0,sigma)
        else:
            coeffs[j] = 0

    return map(lambda x: x.real,  irfft(coeffs * sqrt(N), len(data)))


